import java.awt.Color;
import java.awt.Point;

import jas.ai.ga.*;
import jas.engine.Sim;
import jas.events.ISimEventListener;
import jas.plot.IColored;


/**
 *
 * <p>Title: JAS</p>
 * <p>Description: Java Agent-based Simulation library</p>
 * <p>Copyright (C) 2002 Michele Sonnessa</p>
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * @author Michele Sonnessa
 *
 */
public class Daisy implements IColored, ISimEventListener{

	public static double heatAbsorptionFactor = 20.0;

	public static final int REPRODUCE = 0;
	public static final int CHECK_LIFE = 1;
	
	public static final int DAISY_REPRODUCTION_FACTOR = 4;
	public static final int DAISY_LIVING_FACTOR = -10;

	private int x, y;
	private double albedo;
	private DaisyWorld world;
	private int residualLifeTime, lifeTime;
	
	private double currentPerceivedTemp, currentReproductionCoefficient;
	private double growingLimitFactor, idealTemperature;
	private double mutationProb;
	
	private GARule rule;
	
	public int generatedSeeds = 0;
	
	public double getPerceivedTemperature() { return currentPerceivedTemp;}

	public Daisy(DaisyWorld world, double idealTemperature, double growingLimitFactor, double albedo, int lifeTime, double mutation) {
		this.world = world;
		rule = null;
		this.idealTemperature = idealTemperature;
		this.growingLimitFactor = growingLimitFactor;
		this.albedo = albedo;
		mutationProb = mutation;
		this.lifeTime = lifeTime;
		residualLifeTime = lifeTime;
		
	}
		
	public Daisy(DaisyWorld world, GARule geneticRule, int lifeTime) {
		this.world = world;
		rule = geneticRule;
		growingLimitFactor = GeneticDaisyModel.getGrowFactorFromRule(rule);
		idealTemperature = GeneticDaisyModel.getIdealTemperatureFromRule(rule);
		albedo = GeneticDaisyModel.getAlbedoFromRule(rule);
		mutationProb = 0.0;
		this.lifeTime = lifeTime;
		residualLifeTime = lifeTime;
														
	}

	public void setXY(int x, int y)
	{
		this.x = x;
		this.y = y;
	}

//	public double getHeat(double sunTemperature)
//	{
//		//return sunTemperature * albedo;
//		
//		double solarLuminosity = (world.solarMultiplier + (Math.sin(Sim.getAbsoluteTime() * Math.PI / 365) / 2));
//		double res = Math.pow(solarLuminosity * DaisyWorld.solarFluxConstant * ( 1.0 - albedo ) / DaisyWorld.SBConstant, 0.25) - 273.0;
//		//System.out.println("SUN alb " + albedo + " -> " + res);
//		return res;
//	}

	public void reproduce()
	{
		double localAlbedo = albedo;
		double localTemp = world.temperatureLayer.getDbl(x, y);
		Point[] pts = world.daisyLayer.getMooreNeighborsPositions(x, y);
		for (int i = 0; i < pts.length; i++) {
			Point point = pts[i];
			int xx = (int) pts[i].getX();
			int yy = (int) pts[i].getY();
			Daisy daisy = (Daisy) world.daisyLayer.get(xx, yy);
			if (daisy != null)
				localAlbedo += daisy.getAlbedo();
			else
				localAlbedo += 0.5;
			localTemp += world.temperatureLayer.getDbl(xx, yy);
		}
		localAlbedo /= 9.0;
		localTemp /= 9.0;
		//currentPerceivedTemp = world.getSunTemperature() * albedo;
		
		currentPerceivedTemp = heatAbsorptionFactor * (localAlbedo - albedo ) + localTemp;
		//currentPerceivedTemp = world.temperatureLayer.getDbl(x, y);
		
		currentReproductionCoefficient = 1.0 - Math.pow((idealTemperature - currentPerceivedTemp) / growingLimitFactor, 2.0);
		
		if (currentReproductionCoefficient > 0)
		{		
			int seeds = (int) (DAISY_REPRODUCTION_FACTOR * currentReproductionCoefficient);
			generatedSeeds += seeds;
			//Point[] pts = world.daisyLayer.getMooreNeighborsPositions(x, y);
				
			for (int i = 0; i < seeds; i++)
			{
				int k = Sim.getRnd().getIntFromTo(0, 7);
				int xx = (int) pts[k].getX();
				int yy = (int) pts[k].getY();
				if (world.daisyLayer.get(xx, yy) == null)
				{
					double newAlbedo = albedo;
					if (rule == null && Sim.getRnd().getDblFromTo(0.0, 1.0) < mutationProb)
						newAlbedo = 1.0 - albedo;
					
					Daisy daisy = null;
					if (rule != null)
						//daisy = new Daisy(world, GeneticDaisyModel.masterGa.step(), lifeTime);
						//daisy = new Daisy(world, GeneticDaisyModel.masterGa.step(), lifeTime); 
						daisy = new Daisy(world, rule, lifeTime);
					else
						daisy = new Daisy(world, idealTemperature, growingLimitFactor, newAlbedo, lifeTime, mutationProb);
							
					world.placeMeAt(xx, yy, daisy);
					world.daisies.addSafe(daisy);					
				}
			}
		}
	}

	public void checkLife()
	{
	  boolean isDayToDie = (--residualLifeTime <= 0);
	  
	  int totalLife = lifeTime - residualLifeTime; 
	  
	  if (currentReproductionCoefficient < DAISY_LIVING_FACTOR)
			isDayToDie = true;
			
		if (rule != null)
			rule.setFitness( rule.getFitness() + generatedSeeds);
			//rule.setFitness( generatedSeeds);
			
	  if(isDayToDie)
	  	die();
	}

	public void die()
	{
		world.daisyLayer.set(x, y, null);
		world.daisies.removeSafe(this);
	}

	public Color getColor() {
		
		int temp = (int) Math.max(0, Math.min(albedo * 255.0, 254));
		return new Color(temp, temp, temp);
	}

	/* (non-Javadoc)
	 * @see jas.events.ISimEventListener#performAction(int)
	 */
	public void performAction(int actionId) {
		switch (actionId)
		{
			case REPRODUCE:
				reproduce(); break;
			case CHECK_LIFE:
				checkLife(); break;
		}

	}

	/**
	 * @return
	 */
	public double getAlbedo() {
		return albedo;
	}

	/**
	 * @return
	 */
	public GARule getRule() {
		return rule;
	}

}
