import jas.ai.ga.GARule;
import jas.ai.ga.GeneticAlgorithm;
import jas.engine.*;
import jas.events.*;
import jas.io.ParameterFrame;

/**
  *	A JAS implementation of the DaisyModel. See article for more details about the model.
  *
  *	(c) 2003 - The Turin Cluster Group, LIASES, University of Torino
  * @author Riccardo Boero (boero@econ.unito.it), Michele Sonnessa (sonnessa@di.unito.it)
  * @author Gigi Ferraris (glf313@libero.it), Matteo Morini (morini@labor-torino.it)
  */
public class GeneticDaisyModel extends SimModel
{
	public int daisyPopulation, daisyLifeTime, sunPeriod;

	public int xSize, ySize;
	public double whiteAlbedo , blackAlbedo;
	public double whiteGrowingLimitFactor, whiteIdealTemperature;
	public double blackGrowingLimitFactor, blackIdealTemperature;
	public double mutationProbability;
	public DaisyWorld world;
	public int phase;
	
	public boolean activateGenetic, disasters;
	private GARule[] rules;
	public GeneticAlgorithm masterGa = null;
	public int geneticPopulation;

	public void setParameters()
	{
		ParameterFrame parameterFrame = new ParameterFrame(this, "Daisy Model", path + "DaisyParams.pForm.xml");
		addSimWindow(parameterFrame);
	}

	private void buildActions()
	{
		SimGroupEvent grp = eventList.scheduleGroup(0, 1);
		grp.addEvent(world, "insulate");
		grp.addEvent(world.daisies, Daisy.REPRODUCE);
		grp.addEvent(world.daisies, Daisy.CHECK_LIFE);
		
		eventList.scheduleSimple(daisyLifeTime, daisyLifeTime, this, "spontaneousGermination");
		
		if (disasters)
		{
			eventList.scheduleSimple(sunPeriod * 6, this, "glaciation");
			eventList.scheduleSimple(sunPeriod * 12, this, "serra");
		}	
	}

	public void buildModel()
	{
		world = new DaisyWorld(xSize, ySize, sunPeriod, phase);
		spontaneousGermination();
				
		buildActions();
	}

	private void buildBlackAndWhite()
	{
		for (int i = 0; i < daisyPopulation; i++)
		{ 
			Daisy daisy = null;
			if (Sim.getRnd().getDblFromTo(0.0, 1.0) > 0.5)
				daisy = new Daisy(world, whiteIdealTemperature, whiteGrowingLimitFactor, whiteAlbedo, daisyLifeTime, mutationProbability);
			else
				daisy = new Daisy(world, blackIdealTemperature, blackGrowingLimitFactor, blackAlbedo, daisyLifeTime, mutationProbability);
				
			world.placeMeRandom(daisy);
			world.daisies.add(daisy);							
		}		
	}
	
	public void getBest()
	{
		if (rules == null)
			return;
			
		for (int i = 0; i < rules.length; i++) {
			GARule rule = rules[i];
			masterGa.setReward(rule, rule.getFitness());
		}
		GARule rule = masterGa.getBestRule();
		if (rule == null)
				return;
		double gF = getGrowFactorFromRule(rule);
		double tmp  = getIdealTemperatureFromRule(rule);
		double alb  = getAlbedoFromRule(rule);
		double fit = rule.getFitness();
		
		double gF2 = 0, tmp2 = 0, alb2 = 0, cnt = 0, fit2 = 0;
		rule = masterGa.getMostDiffusedRule();
		if (rule != null)
		{
			gF2 = getGrowFactorFromRule(rule);
			tmp2  = getIdealTemperatureFromRule(rule);
			alb2  = getAlbedoFromRule(rule);
			cnt = rule.getCopies();
			fit2 = rule.getFitness();
		}

		System.out.println("Best @" + Sim.getAbsoluteTime() + " > gF:" + gF + " - t:" + tmp 
														+ " - a:" + alb + " Fit: " + fit + "\n" );
		System.out.println("Most @" + Sim.getAbsoluteTime() + " > gF:" + gF2 + " - t:" + tmp2
														+ " - a:" + alb2 + " Fit: " + fit2 + " - " + cnt + "\n" );
	}
	
	
	private void buildGenetic()
	{
		if (masterGa == null)
		{
			masterGa = new GeneticAlgorithm(geneticPopulation, 24);
			masterGa.setAutoEvolution(false);
		}
		else
		{
			getBest();
			System.out.println("\nNew generation...");
			masterGa.evolve();
		}
		
		rules = new GARule[daisyPopulation];
		for (int i = 0; i < daisyPopulation; i++)
		{ 
			GARule rule = masterGa.step();
			rules[i] = rule;
			
			Daisy daisy = new Daisy(world, rule, daisyLifeTime);
			
			world.placeMeRandom(daisy);
			world.daisies.add(daisy);							
		}		
	}	

	public void glaciation()
	{
		System.out.println("\nGLACIATION\n");
		world.solarMultiplier = 0.7;
	}	

	public void serra()
	{
		System.out.println("\nSERRA\n");
		world.solarMultiplier = 1.8;
	}	

	public void spontaneousGermination()
	{
		if (activateGenetic)
			buildGenetic();
		else	
			buildBlackAndWhite();
	}

	public double getAverageTemp()
	{
		//System.out.println("SUM " + world.temperatureLayer.sum());
		return world.getAverageTemp();
	}
	
	public static double getGrowFactorFromRule(GARule rule)
	{
		return (double) rule.getLongFromTo(0, 7) * 5.0 / 255.0 + 15.0;
	}
	
	public static double getIdealTemperatureFromRule(GARule rule)
	{
		return (double) rule.getLongFromTo(8, 15) * 15.0 / 255.0 + 15.0;
	}	
	
	public static double getAlbedoFromRule(GARule rule)
	{
		return (double) rule.getLongFromTo(16, 23) / 255.0 * 0.8 + 0.1;
	}		
}

